
#include "fft_ears.h"
#include "fft.h"
#include "snd_utils.h"

#define BUMP_INDEX(x) {x++; x %= SAMPLE_BUF_COUNT;}
#define PREV_INDEX(x, n) ((x+(SAMPLE_BUF_COUNT-n))%SAMPLE_BUF_COUNT)
#define BUMP_WHISPER_INDEX(x) {x = (x+1) % WHISPER_BUF_COUNT;}
#define PREV_DELAY_INDEX(x, n) ((x+(DELAY_COUNT-n))%DELAY_COUNT)

#define DEFAULT_FIDDLE SHIFT
#define FLIPPER_DIVISOR 8

#define WHISPER_UP_LEVEL 2.
#define WHISPER_MAX_LEVEL 256.
#define WHISPER_MIN_LEVEL .1



fft_ears::fft_ears(fft_view *v)
{
	bar_view = v;

	InitFiddle();
	InitGraphicStuff();
	InitAudioStuff();
	InitThreadsEtAl();
	Go();
}	

void fft_ears::Go()
{
	resume_thread(proc_thread);
	ears->EnterStream(NULL, true, (void *)this, _listen, NULL, true);
	mouth->EnterStream(NULL, true, (void *)this, _talk, NULL, true);
}

void fft_ears::InitFiddle()
{
	int32 ktr;
	which_fiddle = (fiddle_mode)DEFAULT_FIDDLE;
	real_val = SCROLL_DEFAULT;
	switch (which_fiddle) {
		case SHIFT:
			fiddle_val = 1.0;
			break;
		case FLIPPER:
			fiddle_val = (SCROLL_DEFAULT/SCROLL_RANGE)*
				(SAMPLE_COUNT/FLIPPER_DIVISOR);
			break;
		case WHISPER:
			fiddle_val = 1.0;
		}
	whisper_index = 0;
	for(ktr = 0; ktr < DELAY_COUNT; ktr++)
		memset((char *)&convolve_bufs[ktr], 0, sizeof(float)*SAMPLE_COUNT);
	convolve_index = 0;
}

void fft_ears::InitGraphicStuff()
{
	look_at = LOOK_AT_OUT;
}

void fft_ears::InitAudioStuff()
{
	dac = new BDACStream();
	dac->SetStreamBuffers(STREAM_BUF_SIZE, STREAM_BUF_COUNT);
	dac->EnableDevice(B_ADC_IN, TRUE);
	dac->EnableDevice(B_LOOPBACK, FALSE);
	dac->SetSamplingRate(SRATE);
	
	adc = new BADCStream();
	adc->SetSamplingRate(SRATE);
	adc->SetStreamBuffers(STREAM_BUF_SIZE, STREAM_BUF_COUNT);
	adc->SetADCInput(B_MIC_IN);
	adc->BoostMic(FALSE);

	ears = new BSubscriber("ears");
	mouth = new BSubscriber("mouth");

	ears->Subscribe(adc);
	mouth->Subscribe(dac);
}

void fft_ears::InitThreadsEtAl()
{
	int32 ktr;
	
	for (ktr = 0; ktr < SAMPLE_BUF_COUNT; ktr++) {
		write_sems[ktr] = create_sem(1,"buf write sem");
		read_sems[ktr] = create_sem(0, "buf read sem");
		write_atoms[ktr] = 0;
		read_atoms[ktr] = 0;
	}
	write_index = 0;

	/* We can't finish the dac output until we have three buffers. */
	read_index = 2;

	proc_continue = true;
	proc_thread = spawn_thread(_proc, "whisper proc", 
		B_REAL_TIME_PRIORITY, (void *)this);

	/* Acquire the first write sem */
	if (atomic_add(&write_atoms[write_index], 1) > 0) 
		if (acquire_sem(write_sems[write_index]) < B_NO_ERROR)
			return;
}
	

fft_ears::~fft_ears(void)
{
	int32 ktr;
	int32 ret;
	proc_continue = false;

	mouth->ExitStream();
	ears->ExitStream();

	for (ktr = 0; ktr < SAMPLE_BUF_COUNT; ktr++) {
		delete_sem(write_sems[ktr]);
		delete_sem(read_sems[ktr]);
	}
	
	/* Kick the proc thread (it's probably waiting in 
	 * receive_data).
	 */
	send_data(proc_thread, 0, NULL, 0);
	wait_for_thread(proc_thread, &ret);

	delete ears;
	delete mouth;
	delete dac;
	delete adc;
}

void fft_ears::SetSRate(float val)
{
	dac->SetSamplingRate(val);
	adc->SetSamplingRate(val);
}

void fft_ears::Comb(float *in, int32 count) {}

void fft_ears::Flipper(float *in, int32 count)
{
	int32 ktr, iktr, mktr;
	float hold_up, hold_down;
	float *up, *down;
	int32 half = count/FLIPPER_DIVISOR;

	int32 offset = max_c(fiddle_val,2);
	int32 iter = (half/offset);
	up = &in[1];
	down = &in[count-2];
	/* We only want to step through the first half of each part. */
	offset /= 2;
	
	for (iktr = 0, mktr = 0; iktr < iter; iktr++) {

		for (ktr = 0; ktr < offset; ktr++, mktr++) {
			hold_up = *up;
			hold_down = *down;
			*up = *(up+offset);
			*down = *(down-offset);
			*(up+offset) = hold_up;
			*(down-offset) = hold_down;
			up++;
			down--;

			/* The division may not be integral */ 
			if (mktr >= half)
				return;
		}
		/* Bump the pointers to the next part */
		up += offset;
		down -= offset;
	}
}


#define WHISPER_MULT ((SRATE / 11025) * 2)
void fft_ears::Whisper(float *in, int32 count) 
{
	int32 half = count/2;
	int32 ktr, mktr, dktr;
	float hold;
	memset((char *)aux, 0, sizeof(float)*count);
	
	for (ktr = half-1; ktr >= 1; ktr--) {
		in[ktr] = min_c(50.0,abs(in[ktr]));
		if (abs(in[ktr]) > fiddle_val) {
			for (dktr=1,mktr = ktr*2;mktr < half; mktr *= WHISPER_MULT, dktr++) {
				in[mktr] = in[ktr]/dktr;
				
			}
			for (dktr=1, mktr= ktr/2; mktr > 1; mktr /= WHISPER_MULT, dktr++) {
				aux[mktr] = in[ktr]/dktr;}
		}

		in[SAMPLE_COUNT-1-ktr] = 0.;
	}
	for (ktr = half-1; ktr >= 1; ktr--) 
		in[ktr] += aux[ktr];

}

void fft_ears::Convolve(float *in, int32 count) 
{
	int32 ktr;
	int32 half = count/2;
	float hold[SAMPLE_COUNT];
	in[0] = 0;
	in[SAMPLE_COUNT-1] = 0;
	memcpy((char *)hold, in, (count*sizeof(float)));
	mul_hermitian(convolve_bufs[convolve_index],in, count);

	memcpy((char *)convolve_bufs[convolve_index++], 
		(char *)hold, (count*sizeof(float)));
	convolve_index %= DELAY_COUNT;
}
		
/* harmonic shift of in fft into out fft. */
void fft_ears::Shift(float *buf, int32 count)
{
	int32 ktr;
	int32 new_bin;
	int32 half = count/2;

	if (fiddle_val == 1.0)
		return;

	memset((char *)aux, 0, sizeof(float)*count);

	if (fiddle_val < 1.0) {
		for (ktr = 0; ktr < half; ktr++) {
			new_bin = (ktr*fiddle_val);
			if (ktr != 0 && new_bin == 0) continue;
			aux[new_bin] += buf[ktr];
			aux[count-1-new_bin] = buf[count-1-ktr];
		}
	}
	else {
		for (ktr = half-1; ktr >= 0; ktr--) {
			new_bin = (ktr*fiddle_val);
			if (new_bin > half-1) continue;
			aux[new_bin] += buf[ktr];
			aux[count-1-new_bin] = buf[count-1-ktr];		
		}
	}		 		
	memcpy(buf, aux, count*sizeof(float));

}

#define SPLATTER_SPEED 2
void fft_ears::Splatter(float *buf, int32 count)
{
	int half = count/2;
	int32 ktr;
	for (ktr = 0; ktr < count; ktr++) 
		buf[ktr] = (aux[ktr]*min_c(fiddle_val*2, .99)) + (buf[ktr]*max_c(1.0-fiddle_val, .1));
	for (ktr = half-1; ktr > SPLATTER_SPEED+1; ktr--) {
		aux[ktr] = buf[ktr-SPLATTER_SPEED];
		aux[half+ktr] = buf[half+ktr+SPLATTER_SPEED];
	}

}
	
void fft_ears::Fiddle(float *buf, int32 count)
{
	switch(which_fiddle) {
		case SHIFT:
			Shift(buf, count);
			break;
		case COMB:
			Comb(buf, count);
			break;
		case WHISPER:
			Whisper(buf, count);
			break;
		case FLIPPER:
			Flipper(buf, count);
			break;	
		case CONVOLVE:
			Convolve(buf, count);	
			break;
		case SPLATTER:
			Splatter(buf, count);
			break;
		}
}

bool fft_ears::_listen(void *data, char *sound, size_t count, void *header)
{
	return ((fft_ears *)data)->Listen((int16 *)sound, count/SAMPLE_SIZE);
}

/* Get an adc buffer; count is the number of individual
 * samples (not sample frames).  write_index is the 
 * index number of the fft buffer to write into.
 */
bool fft_ears::Listen(int16 *sound, int32 count)
{
	static int sktr =0;

	/* Convert from stereo to mono. */
	stereo_short2mono_float(sound, adc_hold, count/2);

	 /* Write into the underlap buffer.
	 * We've already acquired the buffer, either in the init or
	 * through the previous Listen() call.
	 * write_index already points to the buffer.
	 */
	memcpy((char *)&samples[write_index][SAMPLE_COUNT/2], 
			(char *)adc_hold, SAMPLE_BUF_SIZE/2);
	
	/* Send the underlap buffer to the proc thread. */
	if (send_data(proc_thread, write_index, NULL, 0) < B_NO_ERROR)
		return false;
		
	/* Now the same thing for the full input buffer. */
	BUMP_INDEX(write_index);

	if (atomic_add(&write_atoms[write_index], 1) > 0) 
		if (acquire_sem(write_sems[write_index]) < B_NO_ERROR)
			return false;

	memcpy((char *)samples[write_index], adc_hold, SAMPLE_BUF_SIZE);

	/* Send this buffer to the proc thread. */
	if (send_data(proc_thread, write_index, NULL, 0) < B_NO_ERROR)
		return false;
		
	/* And now copy into the first half of the next overlap buffer. */
	BUMP_INDEX(write_index);

	if (atomic_add(&write_atoms[write_index], 1) > 0) 
		if (acquire_sem(write_sems[write_index]) < B_NO_ERROR)
			return false;

	memcpy((char *)samples[write_index], 
		(char *)&adc_hold[SAMPLE_COUNT/2], 
		SAMPLE_BUF_SIZE/2);

	return true;
}


int32 fft_ears::_proc(void *obj)
{
	((fft_ears *)obj)->Proc();
}

void fft_ears::Proc(void)
{
	thread_id dummy;
	int32 index;
	
	while (proc_continue) {

		/* Receive the index of the current input buffer. */
		index = receive_data(&dummy, NULL, 0);		
		if (!proc_continue)
			return;	

		/* Transform. */

		do_fft(samples[index], SAMPLE_COUNT, CLOBBER_BIN_COUNT);

		/* Drawing input signal? */
		if (look_at == LOOK_AT_IN)
			SetBars(samples[index], 0);		

		/* Move or scale the buffers and then invert. */
		Fiddle(samples[index], SAMPLE_COUNT);

		/* Drawing output signal? */
		if (look_at == LOOK_AT_OUT) 
			SetBars(samples[index], 0);		

		do_inv_fft(samples[index], SAMPLE_COUNT);

		/* The buffer can now go to the dac thread. */
		if (atomic_add(&read_atoms[index], -1) > 1)
			release_sem(read_sems[index]);
	}
}


bool fft_ears::_talk(void *data, char *sound, size_t count, void *header)
{
	return ((fft_ears *)data)->Talk((int16 *)sound, count/SAMPLE_SIZE);
}

/* The buffers are unwindowed float samples. */
bool fft_ears::Talk(int16 *sound, int32 count)
{
	static int j = 0;
	static bool first_time = true;
	int32 underlap = PREV_INDEX(read_index, 2);
	int32 current = PREV_INDEX(read_index, 1);
	int32 overlap = read_index;
	
	/* We acquire two buffers ahead (read_index is initialized to 2).
	 * When this acquisition succeeds, we know we have three buffers to 
	 * read from.  
	 */
	if (first_time) {
		if (atomic_add(&read_atoms[underlap], 1) > 0) 
			if (acquire_sem(read_sems[underlap]) < B_NO_ERROR)
				return FALSE;	
	}
	
	if (atomic_add(&read_atoms[current], 1) > 0) 
		if (acquire_sem(read_sems[current]) < B_NO_ERROR)
			return FALSE;	

	if (atomic_add(&read_atoms[overlap], 1) > 0) 
		if (acquire_sem(read_sems[overlap]) < B_NO_ERROR)
			return FALSE;	

	/* If this is the first time, we ignore the underlap buffer. */
	mono_float2stereo_short_overlap_cos_window(
		first_time?NULL:&samples[underlap][SAMPLE_COUNT/2],
		samples[current],
		samples[overlap],
		sound,
		count);

	first_time = false;	
	if (atomic_add(&write_atoms[underlap], -1) > 1) 
		if (release_sem(write_sems[underlap]) < B_NO_ERROR)
			return FALSE;

	BUMP_INDEX(read_index);

	if (atomic_add(&write_atoms[current], -1) > 1) 
		if (release_sem(write_sems[current]) < B_NO_ERROR)
			return FALSE;

	BUMP_INDEX(read_index);
	return TRUE;
}

void fft_ears::SetBars(float *buf, int32 start)
{
	int ktr;
	float magn,z,y;
	for (ktr = 0; ktr < FFT_BAR_COUNT; ktr++) {
		y = buf[ktr+start];
		
		z = buf[FFT_BIN_COUNT-ktr-1];
		magn = pow(pow(y, 2.0) + pow(z, 2.0),.95) ;
		bar_buf[ktr] = (magn > FFT_BAR_MAX) ? 
			FFT_BAR_MAX:(char)floor(magn);

	}
	bar_view->SetBars(bar_buf);
}
void fft_ears::SetFiddleMode(fiddle_mode mode) 
{ 
	which_fiddle = mode; 
	SetFiddleValue(real_val);
 	switch (which_fiddle) {
 		case SPLATTER:	
 			memset((char *)aux, 0, SAMPLE_COUNT*(sizeof(float)));
		default:
			break;
	}
}

void fft_ears::SetFiddleValue(float val) 
{
 	real_val = val;
 	switch (which_fiddle) {
 		case CONVOLVE:
			ConvolveFiddleValue(val);
 			break;
  		case SPLATTER:
 			SplatterFiddleValue(val);
			break;
 		case SHIFT:
 			ShiftFiddleValue(val);
 			break;
 		case COMB:
 			CombFiddleValue(val);
 			break;
 		case WHISPER:
 			WhisperFiddleValue(val);
 			break;
 		case FLIPPER:
 			FlipperFiddleValue(val);
 			break; 	}
}

void fft_ears::ShiftFiddleValue(float newValue) 
{
	if (newValue == SCROLL_DEFAULT) {
			fiddle_val = SHIFT_DEFAULT;
			return;
		}
	if (newValue>SCROLL_DEFAULT) {
		fiddle_val = (((newValue-SCROLL_DEFAULT)/SCROLL_UP_RANGE) 
				* SHIFT_UP_RANGE) + SHIFT_DEFAULT;
		return;
	}
	if (newValue < SCROLL_DEFAULT) {
		fiddle_val = (((newValue-SCROLL_DEFAULT)/SCROLL_DOWN_RANGE) 
				* SHIFT_DOWN_RANGE ) + SHIFT_DEFAULT;
		return;		
	}	
}


void fft_ears::FlipperFiddleValue(float newValue)
{
	fiddle_val = (newValue/SCROLL_RANGE)*(SAMPLE_COUNT/FLIPPER_DIVISOR);

}
void fft_ears::ConvolveFiddleValue(float newValue)
{
	fiddle_val = (newValue/SCROLL_RANGE)*(SAMPLE_COUNT/2);

}

#define WHISPER_MAX_THRESH 10.0
void fft_ears::WhisperFiddleValue(float newValue) 
{
	fiddle_val = (newValue/SCROLL_RANGE)*WHISPER_MAX_THRESH;
}

void fft_ears::CombFiddleValue(float val) {}

void fft_ears::SplatterFiddleValue(float val) 
{
	fiddle_val = (val/SCROLL_RANGE);	
}

